##############################################################################
#
# Copyright (c) 2007 Lovely Systems and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
"""
$Id$
"""
__docformat__ = "reStructuredText"


import md5
import random
import sys
import time
import logging
import memcache
import cPickle
import threading
import persistent
import os
import socket
from zope.schema.fieldproperty import FieldProperty
from zope import interface
from interfaces import IMemcachedClient

TLOCAL = threading.local()

log = logging.getLogger('lovely.memcached')

# base namespace for key management
NS = 'lovely.memcached'
# namespace for key timesamps
STAMP_NS = NS + '.stamps'
# namespace for deps
DEP_NS = NS + '.dep'

class Storage(object):
    pass

class MemcachedClient(persistent.Persistent):
    interface.implements(IMemcachedClient)

    defaultNS = FieldProperty(IMemcachedClient['defaultNS'])
    servers = FieldProperty(IMemcachedClient['servers'])
    defaultLifetime = FieldProperty(
        IMemcachedClient['defaultLifetime'])
    trackKeys = FieldProperty(IMemcachedClient['trackKeys'])

    def __init__(self, servers=None, defaultAge=None,
                 defaultNS=None, trackKeys=None):
        if servers is not None:
            self.servers = servers
        if defaultAge is not None:
            self.defaultAge = defaultAge
        if defaultNS is not None:
            self.defaultNS = defaultNS
        if trackKeys is not None:
            self.trackKeys = trackKeys

    def getStatistics(self):
        return self.client.get_stats()


    def _getNS(self, ns, raw):
        if not ns and self.defaultNS:
            if raw:
                ns = str(self.defaultNS)
            else:
                ns = self.defaultNS
        return ns or None

    def set(self, data, key, lifetime=None, ns=None, raw=False,
            dependencies=[]):
        if lifetime is None:
            lifetime = self.defaultLifetime
        ns = self._getNS(ns, raw)
        if not raw:
            data = cPickle.dumps(data)
        elif not isinstance(data, str):
            raise ValueError, data
        log.debug('set: %r, %r, %r, %r' % (key,
                                           len(data), ns,
                                           lifetime))

        bKey = self._buildKey(key, ns, raw=raw)
        if self.client.set(bKey, data, lifetime):
            self._keysSet(key, ns, lifetime)
            self._depSet(bKey, ns, dependencies)
            return bKey
        return None

    def _depSet(self, key, ns, deps):
        for dep in deps:
            depKey = self._buildDepKey(dep, ns)
            keys = self.client.get(depKey)
            if keys is None:
                keys = (key,)
            else:
                keys = keys +  (key,)
            self.client.set(depKey, keys)

    def query(self, key, default=None, ns=None, raw=False):
        ns = self._getNS(ns, raw)
        res = self.client.get(self._buildKey(key, ns, raw=raw))
        if res is None:
            return default
        if raw:
            return res
        return cPickle.loads(res)

    def _buildDepKey(self, dep, ns):
        return DEP_NS + self._buildKey(dep, ns)

    def invalidate(self, key=None, ns=None, raw=False,
                   dependencies=[]):
        ns = self._getNS(ns, raw)
        log.debug('invalidate: %r, %r '% (key, ns))
        if self.trackKeys:
            self.client.delete(self._buildKey((ns, key), STAMP_NS))
        if key is not None:
            self.client.delete(self._buildKey(key, ns, raw))
        for dep in dependencies:
            depKey = self._buildDepKey(dep, ns)
            keys = self.client.get(depKey)
            if keys is not None:
                self.client.delete(depKey)
                for key in keys:
                    self.client.delete(key)

    def invalidateAll(self):
        # notice this does not look at namespaces
        self.client.flush_all()
        key = (tuple(self.servers), self.trackKeys)
        del self._storages()[key]

    def _buildKey(self, key, ns, raw=False):

        """builds a key for key and ns, if key is a persistent
        object its oid is used

        >>> vc1 = MemcachedClient()
        >>> k1 = vc1._buildKey(1, None)

        of course the key is the same for same arguments
        >>> k1 == vc1._buildKey(1, None)
        True

        the key is an md5 digest
        >>> len(k1)
        32

        for different namespaces the keys are different

        >>> vc2 = MemcachedClient()
        >>> k2 = vc2._buildKey(1, u'vc2')
        >>> k2 != k1
        True

        if key has an oid this is taken

        >>> class A: pass
        >>> a = A()
        >>> b = A()
        >>> a._p_oid = "oid of a"
        >>> b._p_oid = "oid of a"
        >>> ka1 = vc1._buildKey(a, 1)
        >>> kb1 = vc1._buildKey(b, 1)
        >>> ka1 == kb1
        True
        >>> b._p_oid = "oid of b"
        >>> kb1 = vc1._buildKey(b, 1)
        >>> ka1 == kb1
        False

        If we set the key to raw we must provide a string
        """
        if raw is True:
            if ns:
                key = ns+key
            if not isinstance(key, str):
                raise ValueError, repr(key)
            return key

        oid = getattr(key, '_p_oid', None)
        if oid is not None:
            key = oid
        if ns is not None:
            m = md5.new(cPickle.dumps((ns, key)))
        else:
            m = md5.new(cPickle.dumps(key))
        return m.hexdigest()

    @property
    def client(self):
#        servers = getattr(self.storage, 'servers', None)
#         if servers is not self.servers:
#             # we have a change in the list of servers
#             self.storage.client = None
#             self.storage.servers = self.servers
        client = getattr(self.storage, 'client', None)
        if client is None:
            client = self._instantiateClient(debug=0)
            self.storage.client = client
        return client

    def _storages(self):
        return TLOCAL.__dict__

    @property
    def storage(self):
        # we use a thread local storage to have a memcache client for every
        # thread.
        key = (tuple(self.servers), self.trackKeys)
        storage = self._storages().get(key)
        if storage is None:
            log.info('Creating new local storage with key %r' % (key,))
            storage = self._storages()[key] = Storage()
        if self.trackKeys and not hasattr(storage, 'keys'):
            tName = threading.currentThread().getName()
            pid = os.getpid()
            hostName = socket.gethostname()
            uid = '%s-%s-%s' % (hostName,
                                pid,
                                tName)
            self._keysInit(storage, uid)
        return storage

    def _instantiateClient(self, debug):
        return memcache.Client(self.servers, debug=debug)

    def _keysInit(self, storage, uid, clients=None):
        log.info('Init of keytracking uid: %r' % uid)
        storage.keys = {}
        storage.uid = uid
        storage.dirtyKeys = set()
        storage.lastUpdates = {}
        if clients is None:
            clients = self._getClients()
        if not storage.uid in clients:
            log.info('Adding new client uid: %r' % storage.uid)
            clients.add(storage.uid)
            self.set(clients, 'clients', lifetime=0, ns=NS)
        return clients

    def _keysSet(self, key, ns, lifetime):
        """track a key"""
        if not self.trackKeys or ns in (NS, STAMP_NS): return
        s = self.storage
        keys = s.keys.get(ns)
        if keys is None:
            keys = set([key])
            s.keys[ns] = keys
        elif not key in keys:
            keys.add(key)
        self.set(s.uid, (ns, key), lifetime=lifetime, ns=STAMP_NS)
        self.set(keys, (s.uid, ns), lifetime=0, ns=NS)

    def _getClients(self):
        return self.query('clients', set(), ns=NS)

    def keys(self, ns=None):
        if not self.trackKeys:
            raise NotImplementedError, "trackKeys not enabled"
        res = set()
        s = self.storage
        t = time.time()
        clients = self._getClients()
        changed = False
        if not s.uid in clients:
            clients = self._keysInit(s, s.uid, clients)
        localKeys = s.keys.get(ns, set())
        for client in clients:
            if client == s.uid:
                v = localKeys
            else:
                v = self.query((client, ns), default=set(), ns=NS)
            res.update(v)
        # look at the timestamps
        for k in list(res):
            uid = self.query((ns, k), ns=STAMP_NS)
            if uid is None:
                if localKeys:
                    changed=True
                    localKeys.discard(k)
                res.discard(k)
        if changed:
            # update the server, we do this just here, because the
            # keys method always looks at the stamps, so it is not
            # required to delete keys from the stored keylist at key
            # setting time
            s.dirtyKeys.add(ns)
            self._keysUpdate(localKeys, ns)
        return res


    def _keysUpdate(self, keys, ns):
        # updates the key set of this thread on server
        s = self.storage
        if not ns in s.dirtyKeys:
            return
        t = time.time()
        # we update only every 5 minutes
        if s.lastUpdates.get(ns, 0) + 300 > t:
            return
        self.set(keys, (s.uid, ns), lifetime=0, ns=NS)
        s.dirtyKeys.discard(ns)
        s.lastUpdates[ns] = t



##############################################################################
#
# Copyright (c) 2007 Lovely Systems and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
"""
$Id$
"""
__docformat__ = "reStructuredText"

import unittest
from zope import component
from zope.testing import doctest
from zope.testing.doctestunit import DocTestSuite, DocFileSuite

from zope.keyreference import testing
from zope.intid.interfaces import IIntIds
from zope.intid import IntIds

from zope.app.testing import setup


def setUp(test):
    test.globs['root'] = setup.placefulSetUp(site=True)
    component.provideUtility(IntIds(), IIntIds)
    component.provideAdapter(testing.SimpleKeyReference)


def tearDown(test):
    setup.placefulTearDown()


def test_suite():
    level1Suites = (
        DocFileSuite(
            'testing/README.txt',
            optionflags=doctest.NORMALIZE_WHITESPACE|doctest.ELLIPSIS,
        ),
        DocTestSuite(
            'lovely.memcached.utility',
            optionflags=doctest.NORMALIZE_WHITESPACE|doctest.ELLIPSIS,
        ),
        )
    level2Suites = (
        DocFileSuite(
            'README.txt',
            setUp=setUp, tearDown=tearDown,
            optionflags=doctest.NORMALIZE_WHITESPACE|doctest.ELLIPSIS,
        ),
        )
    for suite in level2Suites:
        suite.level = 2
    return unittest.TestSuite(level2Suites + level1Suites)

if __name__ == '__main__':
    unittest.main(defaultTest='test_suite')



##############################################################################
#
# Copyright (c) 2006 Lovely Systems and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
"""
$Id$
"""
__docformat__ = 'restructuredtext'

from zope import interface
from zope import component
from zope import event

from zope.intid.interfaces import IIntIds

from interfaces import IInvalidateCacheEvent, IMemcachedClient


class InvalidateCacheEvent(object):
    interface.implements(IInvalidateCacheEvent)

    def __init__(self,
            cacheName=None, key=None, ns=None, raw=False, dependencies=[]):
        self.cacheName = cacheName
        self.key = key
        self.ns = ns
        self.raw = raw
        self.dependencies = dependencies


@component.adapter(IInvalidateCacheEvent)
def invalidateCache(event):
    if event.cacheName is not None:
        cache = component.queryUtility(IMemcachedClient, event.cacheName)
        caches = []
        if cache is not None:
            caches.append(cache)
    else:
        caches = component.getAllUtilitiesRegisteredFor(IMemcachedClient)
    for cache in caches:
        if event.raw is not None:
            cache.invalidate(event.key, event.ns, event.raw, event.dependencies)
        else:
            cache.invalidate(event.key, event.ns, True, event.dependencies)
            cache.invalidate(event.key, event.ns, False, event.dependencies)



##############################################################################
#
# Copyright (c) 2007 Lovely Systems and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
"""
$Id$
"""
__docformat__ = "reStructuredText"

from zope import interface
from zope import schema

class IMemcachedClient(interface.Interface):

    """A memcache client utility"""

    defaultNS = schema.TextLine(
        title=u'Default Namespace',
        description=u"The default namespace used by this client",
        required=False,
        default=None)

    servers = schema.List(
        title = u'Servers',
        description = u"Servers defined as <hostname>:<port>",
        value_type = schema.BytesLine(),
        required = True,
        default=['127.0.0.1:11211']
        )

    defaultLifetime = schema.Int(
        title = u'Default Lifetime',
        description = u'The default lifetime of entries',
        required = True,
        default = 3600,
        )

    trackKeys = schema.Bool(
        title = u'Track Keys',
        description = u'Enable the keys method',
        required = False,
        default = False,
        )

    def getStatistics():
        """returns the memcached stats"""

    def set(data, key, lifetime=None, ns=None,
            raw=False,
            dependencies=[]):
        """Sets data with the given key in namespace. Lifetime
        defaults to defautlLifetime and ns defaults to the
        default namespace.

        The dependencies argument can be used to invalidate on
        specific dependency markers.

        This method returns the key that is generated by the utility.

        If raw is True, the key is taken as is, and is not modified by
        the utility, the key and the namespace need to be strings in
        this case, otherwise a ValueError is raised.
        """

    def query(key, default=None, ns=None, raw=False):
        """query the cache for key in namespace, returns default if
        not found. ns defaults to default namespace."""

    def invalidate(key=None, ns=None, raw=False, dependencies=[]):
        """invalidates key in namespace which defaults to default
        namespace, currently we can not invalidate just a namespace.

        If dependencies are provided all, entries with such
        dependencies are invalidated.
        """

    def invalidateAll():
        """invalidates all data of the memcached servers, not that all
        namespaces are invalidated"""

    def keys(ns=None):
        """if trackKeys is True, returns the keys defined in the
        namespace"""


class IInvalidateCacheEvent(interface.Interface):
    """An event which invalidates cache entries."""

    cacheName = schema.TextLine(
            title = u'cacheName',
            description = u"""
                Invalidate in the cache with this name.
                If no name is given all caches are invalidated.
                """,
            required = False,
            )

    key = schema.TextLine(
            title = u'key',
            required = False,
            )

    ns = schema.TextLine(
            title = u'namespace',
            required = False,
            )

    raw = schema.Bool(
            title = u'raw',
            required = False,
            default = False,
            )

    dependencies = schema.List(
            title = u'Dependencies',
            required = False,
            )



##############################################################################
#
# Copyright (c) 2007 Lovely Systems and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
"""
$Id$
"""
__docformat__ = "reStructuredText"

from datetime import datetime, timedelta

from lovely.memcached.utility import MemcachedClient
import threading

# the own client storage for testing
TLOCAL = threading.local()

class TestMemcachedClient(MemcachedClient):
    """A memcache client which doesn't need a running memcache daemon"""

    def __init__(self, *args, **kw):
        super(TestMemcachedClient, self).__init__(*args, **kw)
        # reset counts etc
        self.resetCounts()

    def _instantiateClient(self, debug):
        return SimulatedMemcached()

    def _storages(self):
        return TLOCAL.__dict__

    @property
    def hits(self):
        return self.client.hits

    @property
    def misses(self):
        return self.client.misses

    def resetCounts(self):
        self.client.resetCounts()


class SimulatedMemcached(object):

    def __init__(self):
        self.cache = {}
        self.resetCounts()

    def getStats(self):
        return []

    def set(self, key, data, lifetime=0):
        # raise an error if not a string
        str(key)
        str(data)
        if lifetime:
            lifetime = datetime.now()+timedelta(seconds=lifetime)
        else:
            lifetime = None
        self.cache[key] = (data, lifetime)
        return True

    def get(self, key):
        str(key)
        data = self.cache.get(key, None)
        if data is None:
            self._misses += 1
            return None
        if data[1] is None or datetime.now()<data[1]:
            self._hits += 1
            return data[0]
        del self.cache[key]
        self._misses += 1
        return None

    def delete(self, key):
        if key in self.cache:
            del self.cache[key]

    def flush_all(self):
        self.cache = {}

    @property
    def hits(self):
        return self._hits

    @property
    def misses(self):
        return self._misses

    def resetCounts(self):
        self._hits = 0
        self._misses = 0



##############################################################################
#
# Copyright (c) 2007 Lovely Systems and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
"""
$Id$
"""
__docformat__ = "reStructuredText"

import interfaces
import utility
from zope import component
from zope.interface.interfaces import IMethod
from zope import schema
from z3c.configurator import configurator
from zope.component.interfaces import ISite
from zope.lifecycleevent import ObjectCreatedEvent
import zope.event
from zope.security.proxy import removeSecurityProxy

class IMemcachedClientProperties(interfaces.IMemcachedClient):

    name = schema.TextLine(title=u'Name',
                           required=False,
                           default=u'')

class SetUpMemcachedClient(configurator.SchemaConfigurationPluginBase):
    component.adapts(ISite)
    schema = IMemcachedClientProperties

    def __call__(self, data):

        for name in self.schema:
            field = self.schema[name]
            if IMethod.providedBy(field):
                continue
            if data.get(name) is None:
                data[name] = self.schema[name].default
        name = data.get('name')
        site = self.context
        # we just wanna have one
        util = component.queryUtility(interfaces.IMemcachedClient,
                                      context=site,
                                      name=name)
        if util is not None:
            return
        # needed to be called TTW
        sm = removeSecurityProxy(site.getSiteManager())
        default = sm['default']
        util = utility.MemcachedClient()
        zope.event.notify(ObjectCreatedEvent(util))
        for attr in ['trackKeys', 'defaultNS', 'servers',
                     'defaultLifetime']:
            setattr(util, attr, data[attr])
        default[u'memcachedclient_' + name] = util
        sm.registerUtility(util, interfaces.IMemcachedClient,
                           name=name)



##############################################################################
#
# Copyright (c) 2007 Lovely Systems and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
"""
$Id$
"""
__docformat__ = "reStructuredText"

import os.path
import unittest
from zope.testing import doctest
from zope.app.testing import functional
from zope.app.testing.functional import ZCMLLayer

memcachedLayer = ZCMLLayer(
    os.path.join(os.path.split(__file__)[0], 'ftesting.zcml'),
    __name__, 'memcachedLayer', allow_teardown=True)


def test_suite():
    suite = functional.FunctionalDocFileSuite('README.txt')
    suite.layer = memcachedLayer
    suite.level = 2
    return unittest.TestSuite((suite,))

if __name__ == '__main__':
    unittest.main(defaultTest='test_suite')



#!/usr/bin/env python
""" Installer
"""
import os
from setuptools import setup, find_packages

NAME = 'lovely.memcached'
PATH = ['src'] + NAME.split('.') + ['version.txt']
VERSION = open(os.path.join(*PATH)).read().strip()

setup(
    name=NAME,
    version=VERSION,
    author="Lovely Systems",
    author_email="office@lovelysystems.com",
    description="A memcached client utiltiy for zope 3",
    long_description=(open("README.txt").read() + "\n" +
                      open("CHANGES.txt").read()),
    license="ZPL 2.1",
    keywords="zope3 zope memcached cache ram",
    url='http://github.com/lovelysystems/lovely.memcached',
    packages=find_packages('src'),
    include_package_data=True,
    package_dir={'':'src'},
    namespace_packages=['lovely'],
    extras_require=dict(test = ['z3c.configurator',
                                'zope.keyreference',
                                'zope.securitypolicy',
                                'zope.app.testing',
                                'zope.testbrowser',
                                ]),
    install_requires = ['setuptools',
                        'python-memcached',
                        'zope.site',
                        'zope.intid',
                        'zope.event',
                        'zope.interface',
                        'zope.lifecycleevent',
                        'zope.schema',
                        'zope.security'
                        ],
    zip_safe = False,
)
